/*
 * Copyright (C) 2007 Ogochan.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/*
 * to compile
 *
 * gcc raidzap.c -o raidzap
 */

#define		_LARGEFILE64_SOURCE 
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<linux/types.h>
#include	<linux/raid/md_p.h>
#include	<libgen.h>
#include	<string.h>

typedef	unsigned int		u32;
typedef	unsigned long long	u64;

/*
 *	This function came from linux/drivers/md/md.c
 */
static	unsigned	int
calc_sb_csum(
	mdp_super_t	*sb)
{
	u64		newcsum;
	u32		*sb32 = (u32 *)sb;
	int		i;
	unsigned int	disk_csum
		,			csum;

	newcsum = 0;
	disk_csum = sb->sb_csum;
	sb->sb_csum = 0;

	for ( i = 0 ; i < MD_SB_BYTES / 4 ; i++ ) {
		newcsum += sb32[i];
	}
	csum = (newcsum & 0xffffffff) + (newcsum>>32);


	sb->sb_csum = disk_csum;

	return	(csum);
}

static	char	*
dev_state(
	__u32	state)
{
	static	char	buff[5+1];
	char	*p;

	p = buff;
	*p ++ = ( state + '0' );
	if		(  ( state & ( 1 << MD_DISK_FAULTY ) )  !=  0  ) {
		*p ++ = 'F';
	}
	if		(  ( state & ( 1 << MD_DISK_SYNC ) )  !=  0  ) {
		*p ++ = 'S';
	}
	if		(  ( state & ( 1 << MD_DISK_REMOVED ) )  !=  0  ) {
		*p ++ = 'R';
	}
	if		(  ( state & ( 1 << MD_DISK_ACTIVE ) )  !=  0  ) {
		*p ++ = 'A';
	}
	*p = 0;

	return	(buff);
}

static	void
do_info(
	char	*dev)
{
	int		fd;
	char	buff[MD_SB_BYTES];
	mdp_super_t	*super;
	off64_t	x
		,	y;
	int		i;

	if		(  ( fd = open(dev,O_RDONLY) )  >  0  ) {
		x = lseek64(fd,0,SEEK_END);
		y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES;
		lseek64(fd,y,SEEK_SET);
		read(fd,buff,MD_SB_BYTES);
		super = (mdp_super_t *)buff;
		if		(  super->md_magic  ==  MD_SB_MAGIC  ) {
			printf("Yes MD\n");
			printf("version      = %d.%d.%d\n",
				   super->major_version, super->minor_version, super->patch_version);
			printf("uuid         = %x-%x-%x-%x\n",
				   super->set_uuid0, super->set_uuid1, super->set_uuid2, super->set_uuid3);
			printf("event_lo     = %d\n",super->events_lo);
			printf("event_hi     = %d\n",super->events_hi);
			printf("level        = %d\n",super->level);
			printf("nr_disks     = %d\n",super->nr_disks);
			printf("active_disks = %d\n",super->active_disks);
			printf("spare_disks  = %d\n",super->spare_disks);
			printf("failed_disks = %d\n",super->failed_disks);
			printf("chunk_size   = %d\n",super->chunk_size);
			printf("state        = %s\n", ( super->state == MD_SB_CLEAN ) ? "clean" : "errors");
			printf("this state   = %s\n",dev_state(super->this_disk.state));
			printf("disk number  = %d\n",super->this_disk.number);
			i = super->this_disk.number;
			printf("  no        = %d\n",super->disks[i].number);
			printf("  dev       = %d,%d\n",super->disks[i].major, super->disks[i].minor);
			printf("  raid_disk = %d\n",super->disks[i].raid_disk);
			printf("  state     = %s\n",dev_state(super->disks[i].state));
			for	( i = 0 ; i < super->raid_disks; i ++ ) {
				printf("    no        = %d\n",super->disks[i].number);
				printf("    dev       = %d,%d\n",super->disks[i].major, super->disks[i].minor);
				printf("    raid_disk = %d\n",super->disks[i].raid_disk);
				printf("    state     = %s\n",dev_state(super->disks[i].state));
			}
		} else {
			printf("No this device is not md %s\n",dev);
		}
		close(fd);
	} else {
		printf("device not found\n");
	}
}

/*
 * CAUTION !!!!!!!!!!
 * If you can't understand this function, YOU DON'T USE THIS PROGRAM.
 */
static	void
//do_repair(void)	
do_repair(
	char	*devt,
	char	*dev)
{
	int		fd
		,	fdt;
	char	buff[MD_SB_BYTES];
	char	bufft[MD_SB_BYTES];
	mdp_super_t	*super;
	off64_t	x
		,	y;
	int		i;
//	char	*dev = "/dev/sdk1";		/*	repair disk		*/
//	char	*devt = "/dev/sdg1";	/*	other safe disk	*/
	__u32	events_hi;
	__u32	events_lo;

	if		(  ( fdt = open(devt,O_RDONLY) )  >  0  ) {
		x = lseek64(fdt,0,SEEK_END);
		y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES;
		lseek64(fdt,y,SEEK_SET);
		read(fdt,bufft,MD_SB_BYTES);
		close(fdt);
		super = (mdp_super_t *)bufft;
		events_hi = super->events_hi;
		events_lo = super->events_lo;
	} else {
		printf("can't read master %s\n",devt);
		exit(1);
	}
	if		(  ( fd = open(dev,O_RDWR) )  >  0  ) {
		x = lseek64(fd,0,SEEK_END);
		y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES;
		lseek64(fd,y,SEEK_SET);
		if		(  read(fd,buff,MD_SB_BYTES)  ==  MD_SB_BYTES  ) {
			if		(  super->md_magic  ==  MD_SB_MAGIC  ) {
				//super = (mdp_super_t *)buff;
				super->this_disk.state = 6;
				i = super->this_disk.number = 6;
				super->disks[i].raid_disk = 3;
				super->disks[i].number = 3;
				super->disks[i].state = 6;
				super->events_hi = events_hi;
				super->events_lo = events_lo;
				super->sb_csum = calc_sb_csum(super);
				lseek64(fd,y,SEEK_SET);
				write(fd,bufft,MD_SB_BYTES);
			} else {
				printf("No this device is not md \n");
			}
		} else {
			printf("can't read superblock %s\n",dev);
		}
	} else {
		printf("can't open device %s\n",dev);
	}
	close(fd);
}

static	void
do_save(
	char	*dev)
{
	int		fd
		,	fdb;
	char	buff[MD_SB_BYTES]
		,	name[1024];
	mdp_super_t	*super;
	off64_t	x
		,	y;

	if		(  ( fd = open(dev,O_RDWR) )  >  0  ) {
		x = lseek64(fd,0,SEEK_END);
		y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES;
		lseek64(fd,y,SEEK_SET);
		if		(  read(fd,buff,MD_SB_BYTES)  ==  MD_SB_BYTES  ) {
			super = (mdp_super_t *)buff;
			if		(  super->md_magic  ==  MD_SB_MAGIC  ) {
				sprintf(name,"%s.bak",basename(dev));
				if		(	(  ( fdb = open(name,O_WRONLY|O_CREAT) )  >  0  )
						&&	(  write(fdb,buff,MD_SB_BYTES)  ==  MD_SB_BYTES  ) ) {
					close(fdb);
				} else {
					printf("can't write superblock backup\n");
				}
			} else {
				printf("No this device is not md %s\n",dev);
			}
		} else {
			printf("can't read superblock\n");
		}
	} else {
			printf("can't open device\n");
	}
	close(fd);
}

extern	int
main(
	int		argc,
	char	**argv)
{
printf("argc %d\n",argc);
	if (argc < 3) {
		printf("raidzap -l/-s DEVICE\n");
		printf("raidzap -r   MASTER REPARE\n");
	}else{
		if		(  strcmp(argv[1],"-l")  ==  0  ) {
			printf("INFO:%s\n",argv[2]);
			do_info(argv[2]);
		}
		if		(  strcmp(argv[1],"-s")  ==  0  ) {
			printf("SAVE:%s\n",argv[2]);
			do_save(argv[2]);
		}

#if	1
	/*	If you want to repair your md, you should change do_repair()
		 *	and un-comment this. */
		if		(  strcmp(argv[1],"-r")  ==  0  ) {
			if (argc < 4) {
				printf("raidzap -r   MASTER REPARE\n");
			}else{
				printf("repair:Master:%s Repair:%s\n",argv[2],argv[3]);
				do_repair(argv[2],argv[3]);
			}
		}
#endif
	}
}
